import functools
import multiprocessing
import pathlib

from PIL import Image, ImageDraw, ImageFont
from tqdm import tqdm

from .commons import RE_DIGITS, LABEL_SELECTED, LABEL_DISCARDED


def stem_file_list(selected_files):
    return {pathlib.Path(filename).stem for filename in selected_files}


def annotate_color_coded_images(file_stems, parameters):
    colors = {LABEL_DISCARDED: (198, 78, 82), LABEL_SELECTED: (74, 114, 174)}
    roi_rectangle = [
        (parameters.roi_left, parameters.roi_top),
        (parameters.roi_right, parameters.roi_bottom),
    ]
    font_y_pos = parameters.image_height - 25 - 18
    color_coded_images = [
        p
        for p in parameters.colored_dir.iterdir()
        if not p.name.startswith(".")
    ]
    func = functools.partial(
        annotate_one_color_coded_image,
        file_stems=file_stems,
        roi_rectangle=roi_rectangle,
        colors=colors,
        font_y_pos=font_y_pos,
    )
    cpu_count = multiprocessing.cpu_count()
    total = len(color_coded_images)
    with multiprocessing.Pool(cpu_count) as mpp:
        list(tqdm(mpp.imap(func, color_coded_images), total=total))


def annotate_one_color_coded_image(
    path, file_stems, roi_rectangle, colors, font_y_pos
):
    # fonts can't be pickled, we need to open it every time.
    try:
        font = ImageFont.truetype("Courier New.ttf", 18)
    except OSError:
        font = None
    label = LABEL_SELECTED if path.stem in file_stems else LABEL_DISCARDED
    new_path = path.with_name(f"{path.stem}_{label}{path.suffix}")
    img = Image.open(path)
    # draw a colored roi
    draw = ImageDraw.Draw(img)
    color = colors[label]
    draw.rectangle(roi_rectangle, outline=color, width=2)
    # add sequence number
    sequence_nr = RE_DIGITS.search(path.name).group()
    draw.text((25, font_y_pos), sequence_nr, font=font)
    img.save(new_path)
    path.unlink()


def rename_cut_images(file_stems, parameters):
    for path in tqdm(list(parameters.cuts_dir.iterdir())):
        if not path.stem.startswith("."):
            label = (
                LABEL_SELECTED if path.stem in file_stems else LABEL_DISCARDED
            )
            new_path = path.with_name(f"{path.stem}_cut_{label}{path.suffix}")
            path.rename(new_path)
