import multiprocessing
import functools

from tqdm import tqdm

from .commons import WorkflowResult
from .parameters import Parameters
from .prescan import scan_tifs, prepare_output_dirs
from .imageproc import process_one_tif
from .dataproc import (
    set_plotting_styles,
    construct_data_frame,
    find_guard_threshold,
    check_guards,
    select_on_guards,
    smooth_rolling_min,
    smooth_savgol_filter,
    find_extremas,
    save_data,
    create_report,
)
from .postproc import (
    stem_file_list,
    annotate_color_coded_images,
    rename_cut_images,
)


def prescan_workflow(folder, top, right, bottom, left, **kargs):
    print("1/5: scanning tifs for common autocontrast values")

    parameters = Parameters(folder, top, right, bottom, left, **kargs)
    parameters = scan_tifs(parameters)

    parameters.offset = parameters.intensity_min
    max_offset = parameters.intensity_max - parameters.intensity_min
    parameters.scale = max_offset / 255 / parameters.boost

    prepare_output_dirs(parameters)

    return WorkflowResult(None, parameters)


def image_workflow(parameters):
    print("2/5: Image analysis and conversion")
    func = functools.partial(process_one_tif, parameters=parameters)
    tif_list = parameters.tif_list
    num_items = len(tif_list)
    cpu_count = multiprocessing.cpu_count()
    with multiprocessing.Pool(cpu_count) as mpp:
        stat_results = list(tqdm(mpp.imap(func, tif_list), total=num_items))
    return WorkflowResult(stat_results, parameters)


def data_workflow(stats_results, parameters):
    print("3/5: Data analysis")
    set_plotting_styles()
    data_frame = construct_data_frame(stats_results, parameters)

    # uncomment the next two lines to use the main run_test_data
    # from .dataproc import save_temp
    # save_temp(data_frame, parameters)

    find_guard_threshold(data_frame, parameters)
    data_frame = check_guards(data_frame, parameters)

    selected_df = select_on_guards(data_frame, parameters)
    selected_df = smooth_rolling_min(selected_df, parameters)
    selected_df = smooth_savgol_filter(selected_df, parameters)

    extremas_df = find_extremas(selected_df, parameters)

    save_data(data_frame, selected_df, extremas_df, parameters)
    create_report(data_frame, selected_df, extremas_df, parameters)

    return WorkflowResult(list(selected_df["file"]), parameters)


def postprocessing_workflow(selected_files, parameters):
    print("4/5: Post processing, renaming cut images")
    file_stems = stem_file_list(selected_files)
    rename_cut_images(file_stems, parameters)
    print("5/5: Post processing, annotating color coded images")
    annotate_color_coded_images(file_stems, parameters)
    return WorkflowResult(None, parameters)


def cached_data_workflow(folder):
    from .dataproc import load_temp

    print("3/5: Data analysis")
    set_plotting_styles()
    data_frame, parameters = load_temp(folder)
    find_guard_threshold(data_frame, parameters)
    data_frame = check_guards(data_frame, parameters)

    selected_df = select_on_guards(data_frame, parameters)
    selected_df = smooth_rolling_min(selected_df, parameters)
    selected_df = smooth_savgol_filter(selected_df, parameters)

    extremas_df = find_extremas(selected_df, parameters)

    save_data(data_frame, selected_df, extremas_df, parameters)
    create_report(data_frame, selected_df, extremas_df, parameters)

    return WorkflowResult(list(selected_df["file"]), parameters)
